#include "data.h"

extern PPROJECT ActiveProject;
extern MYPTR HighItem;
extern TYPE HighType;

edge* CreateEdge(PMODEL m,vert* a,vert* b,BOOL make_delta) {
	edge* e;

	if(m==NULL) m=(PMODEL)ActiveProject;

	e=(edge*)malloc(sizeof(edge));
	if(e!=NULL) {
//		memset(e,0,sizeof(EDGE));
		addTail(&m->Edges,(link*)e);
		e->VertS=a;
		e->VertE=b;
		e->FaceL=e->FaceR=0;
		e->Flags=EF_NOLEFT|EF_NORIGHT;
		e->Index=m->NextEdgeIndex++;
		if(make_delta) CreateEdgeDelta(e,DT_ADD);
	}
	return e;
}

PDELT CreateEdgeDelta(edge* e,TYPE type) {
	PMODEL m;
	PDELT d;
	LSMWORD extra[2];

	m=FindModel((link*)e);

	d=CreateDelta((PPROJECT)m,type,EDGEDELT,0);
	if(d==NULL) {
		ReportSystemError("Creation of Delta object failed.");
		return d;
	}
	d->Index=e->Index;

	if(type==DT_DEL) return d;

	extra[0].l=e->VertS->Index;
	extra[1].l=e->VertE->Index;

	ExtendDelta(d,2,extra);

	if(e->Flags&TF_SHARP) {
		extra[0].l=SHARPFLAG;
		ExtendDelta(d,1,extra);
	}

	return d;
}

void DestroyEdge(edge* e,BOOL make_delta) {
	PMODEL m;
	face *f,*nf;

	if(e==NULL) return;

	m=FindModel((link*)e);

	f=FIRSTFACE(ActiveProject);
	while((face*)NEXT(f)) {
		nf=(face*)NEXT(f);
		if(f->EdgeE==e || f->EdgeF==e || f->EdgeG==e || f->EdgeH==e)
		DestroyFace(f,make_delta);
		f=nf;
	}

	if(make_delta) CreateEdgeDelta(e,DT_DEL);

	if(HighItem.e==e) {
		HighItem.e=NULL;
		HighType=HT_NOTHING;
	}

	((PPROJECT)m)->Flags|=MF_MESHCHANGED;
	remLink((link*)e);
	free(e);

	return;
}

edge* FindEdge(PMODEL m,vert* a,vert* b) {
	edge* e;

	if(m==NULL) m=(PMODEL)ActiveProject;
	STEPEDGES(m,e)
	if((e->VertS==a && e->VertE==b) || (e->VertS==b && e->VertE==a))
	return e;

	return NULL;
}

void FinishEdgeByDelta(PMODEL m,edge* e,PDELT d) {
	UINT i;

	e->Index=d->Index;
	e->VertS=IndexVert(m,d->Data[0].l);
	e->VertE=IndexVert(m,d->Data[1].l);

	i=2;
	while(i<d->Len)
	switch(d->Data[i].l) {
		case SHARPFLAG:
			e->Flags|=TF_SHARP;
			i++;
		break;
		default: i++;
	}
}

edge* GetEdge(PMODEL m,vert* a,vert* b,BOOL make_delta) {
	edge* e;

	if (a==b) return NULL;

	e=FindEdge(m,a,b);
	if(e==NULL) e=CreateEdge(m,a,b,make_delta);
	return e;
}

edge* FindDuplicate(PMODEL m,edge* e) {
	edge* ne;

	if(m==NULL) m=(PMODEL)ActiveProject;

	STEPEDGES(m,ne)
	if(ne!=e) {
		if(e->VertS==ne->VertS && e->VertE==ne->VertE) return ne;
		if(e->VertS==ne->VertE && e->VertE==ne->VertS) return ne;
	}

	return NULL;
}

