#include "data.h"
#include "lsmres.h"
#include <math.h>

extern PPROJECT ActiveProject;
extern MYPTR HighItem;
extern short HighType;
extern HINSTANCE hInst;

void AttachMarked(part* p) {
	vert* v;

	STEPVERTS(ActiveProject,v)
	if(v->Flags&TF_MARKED) {
		AttachVertex(v,p);
//		CreateVertDelta(v,DT_CHG);
	}
//	AddAtomDelta(ActiveProject);
	ClearAll();
}

void AttachVertex(vert* v,part* p) {

	v->Owner=p;
	if(Distance2(v->Spot,p->Base->Spot) > Distance2(v->Spot,p->End->Spot)) {
		v->Flags&=~VF_WITHB;
	} else {
		v->Flags|=VF_WITHB;
	}
//	CreateVertDelta(v,DT_CHG);
//	AddAtomDelta(ActiveProject);
}

void DeleteEdges(BOOL make_delta) {
	edge* e,*ne;

	e=FIRSTEDGE(ActiveProject);
	while(NEXT(e)!=NULL) {
		ne=(edge*)NEXT(e);
		if (e->VertS->Flags & e->VertE->Flags & TF_MARKED) {
			DestroyEdge(e,FALSE);//make_delta);
		}
		e=ne;
	}
//	AddAtomDelta(ActiveProject);
	InvalidateView();
}

void DeleteFaces(BOOL make_delta) {
	face *f,*nf;
	FLAGS q;

	f=FIRSTFACE(ActiveProject);
	while(NEXT(f)!=NULL) {
		nf=(face*)NEXT(f);
		q=TF_MARKED & f->VertA->Flags & f->VertB->Flags & f->VertC->Flags;
		if(f->VertD) q&=f->VertD->Flags;
		if(q) DestroyFace(f,FALSE);//make_delta);
		f=nf;
	}
//	AddAtomDelta(ActiveProject);
	InvalidateView();
}

void DeleteHighlight(BOOL verify) {

	switch(HighType) {
		case HT_VERT:
			DestroyVertex(HighItem.v,FALSE);
		break;
		case HT_PART:
			DestroyPart(HighItem.p,FALSE);
		break;
		case HT_KNOB:
			DestroyKnob(HighItem.k,FALSE);
		break;
		case HT_EDGE:
			DestroyEdge(HighItem.e,FALSE);
		break;
		case HT_FACE:
			DestroyFace(HighItem.f,FALSE);
		break;
	}
	HighItem.e=NULL;
	HighType=HT_NOTHING;
//	AddAtomDelta(ActiveProject);
	InvalidateView();
}

void DeleteVerts(BOOL make_delta) {
	vert* v,*nv;

	v=FIRSTVERT(ActiveProject);
	while(NEXT(v)!=NULL) {
		nv=(vert*)NEXT(v);
		if (v->Flags&TF_MARKED) {
			DestroyVertex(v,FALSE);//make_delta);
		}
		v=nv;
	}
//	AddAtomDelta(ActiveProject);
	InvalidateView();
}

void DetachHighlight(BOOL verify) {
	knob *k;

	switch(HighType) {
		case HT_NOTHING:
		break;
		case HT_VERT:
			HighItem.v->Owner=NULL;
			HighItem.v->Flags&=~VF_WITHB;
//			CreateVertDelta(HighItem.v,DT_CHG);
		break;
		case HT_PART:
			HighItem.p->Parent=NULL;
			k=CopyKnob(HighItem.p->Base,FALSE);
			k->Flags=KF_ISBASE;
//			CreateKnobDelta(k,DT_ADD);
			HighItem.p->Base=k;
//			CreatePartDelta(HighItem.p,DT_CHG);
		break;
	}
//	AddAtomDelta(ActiveProject);
}

void EditHighlight(HWND hwnd) {
//	short i;
	MYPTR m;
	matrix qm;
	PDLGX dd;

	if(ActiveProject==NULL) return;

	m.e=HighItem.e;

	switch(HighType) {
	case HT_VERT:
		dd=(PDLGX)malloc(3*sizeof(DLGX));
		dd[0].item=2;
		dd[1].type=DXT_TTL;
		dd[1].data.tdata="Vertex coordinates";
		dd[2].type=DXT_VEC;
		dd[2].item=EB_VECX;
		dd[2].data.vdata=m.v->Spot;
		DialogBoxParam(hInst,MAKEINTRESOURCE(DLG_VECTOR),hwnd,
			InputMatrixDlg,(int)dd);
		DestroyDialogData(dd);
//		CreateVertDelta(m.v,DT_CHG);
	break;
	case HT_UVPT:
		dd=(PDLGX)malloc(3*sizeof(DLGX));
		dd[0].item=2;
		dd[1].type=DXT_NUM;
		dd[1].item=EB_VECX;
		dd[1].data.fdata=m.u->Spot[U];
		dd[2].type=DXT_NUM;
		dd[2].item=EB_VECY;
		dd[2].data.fdata=m.u->Spot[V];
		DialogBoxParam(hInst,MAKEINTRESOURCE(DLG_UVMAP),hwnd,InputMatrixDlg,(int)dd);
		m.u->Spot[U]=dd[1].data.fdata;
		m.u->Spot[V]=dd[2].data.fdata;
		DestroyDialogData(dd);
//		CreateUVPtDelta(m.u,DT_CHG);
	break;
	case HT_PART:
		PrepSingleKnobDrag(m.p->End);
		getInheritedMatrix(m.p,qm);
		Cart2Polar(m.p,qm);

		dd=(PDLGX)malloc(11*sizeof(DLGX));
		dd[0].item=10;
		dd[1].type=DXT_NUM;
		dd[1].item=EB_VECXX;
		dd[1].data.fdata=m.p->AngleX.use;
		dd[2].type=DXT_NUM;
		dd[2].item=EB_VECYX;
		dd[2].data.fdata=m.p->AngleY.use;
		dd[3].type=DXT_NUM;
		dd[3].item=EB_VECZX;
		dd[3].data.fdata=m.p->AngleZ.use;
		dd[4].type=DXT_NUM;
		dd[4].item=EB_VECLX;
		dd[4].data.fdata=m.p->Length.use;
		dd[5].type=DXT_TXT;
		dd[5].item=EB_NAME;
		dd[5].data.tdata=CloneString(m.p->Name);

		dd[6].type=DXT_FLG;
		dd[6].item=CB_LOCKX;
		dd[6].data.idata=(m.p->Flags&PF_LOCKX);
		dd[7].type=DXT_FLG;
		dd[7].item=CB_LOCKY;
		dd[7].data.idata=(m.p->Flags&PF_LOCKY);
		dd[8].type=DXT_FLG;
		dd[8].item=CB_LOCKZ;
		dd[8].data.idata=(m.p->Flags&PF_LOCKZ);
		dd[9].type=DXT_FLG;
		dd[9].item=CB_LOCKL;
		dd[9].data.idata=(m.p->Flags&PF_LOCKL);
		dd[10].type=DXT_OBJ;
//		dd[10].item=CB_LOCKL;
		dd[10].data.mdata.p=m.p;

		DialogBoxParam(hInst,MAKEINTRESOURCE(DLG_EDITPART),hwnd,EditPartDlg,(int)dd);

		DestroyString(m.p->Name);
		m.p->Name=CloneString(dd[5].data.tdata);

		qm[LX]=m.p->Base->Spot[X];
		qm[LY]=m.p->Base->Spot[Y];
		qm[LZ]=m.p->Base->Spot[Z];

		m.p->AngleX.use=dd[1].data.fdata;
		m.p->AngleY.use=dd[2].data.fdata;
		m.p->AngleZ.use=dd[3].data.fdata;
		m.p->Length.use=dd[4].data.fdata;

		m.p->Flags&=~(PF_LOCKX|PF_LOCKY|PF_LOCKZ|PF_LOCKL);
		if(dd[6].data.idata) m.p->Flags|=PF_LOCKX;
		if(dd[7].data.idata) m.p->Flags|=PF_LOCKY;
		if(dd[8].data.idata) m.p->Flags|=PF_LOCKZ;
		if(dd[9].data.idata) m.p->Flags|=PF_LOCKL;

		Polar2Cart(m.p,qm);
		MoveKnobEnd(m.p->End);
		RebuildPartMatrices();
//		CreatePartDelta(m.p,DT_CHG);
		DestroyDialogData(dd);
	break;
	case HT_KNOB:
		dd=(PDLGX)malloc(3*sizeof(DLGX));
		dd[0].item=2;
		dd[1].type=DXT_TTL;
		dd[1].data.tdata="Knob location";
		dd[2].type=DXT_VEC;
		dd[2].item=EB_VECX;
		dd[2].data.vdata=m.k->Spot;

		PrepSingleKnobDrag(m.k);
		if(DialogBoxParam(hInst,MAKEINTRESOURCE(DLG_VECTOR),hwnd,
			InputMatrixDlg,(int)dd)==IDOK) {
			Constrain(m.k);
			MoveKnobEnd(m.k);
			RebuildPartMatrices();
			InvalidateView();
		}
//		CreateKnobDelta(m.p,DT_CHG);
		DestroyDialogData(dd);
	break;
	}
//	AddAtomDelta(ActiveProject);
	InvalidateView();
}

void MarkAll(void) {
	vert* v;

	STEPVERTS(ActiveProject,v)
	v->Flags|=TF_MARKED;

	InvalidateView();
}

void ClearAll(void) {
	vert* v;

	if(ActiveProject==NULL) return;

	STEPVERTS(ActiveProject,v)
	v->Flags&=~TF_MARKED;

	InvalidateView();
}

void SetGridSnap(HWND hwnd) {
	PDLGX dd;

	if(ActiveProject==NULL) return;

	dd=(PDLGX)malloc(2*sizeof(DLGX));
	dd[0].item=1;
	dd[1].type=DXT_NUM;
	dd[1].item=EB_VECX;
	switch(ActiveProject->EditMode) {
		case EM_MESHEDIT:
			dd[1].data.fdata=((PMODEL)ActiveProject)->MeshSnap;
		break;
		case EM_MAPEDIT:
			dd[1].data.fdata=((PMODEL)ActiveProject)->MapSnap;
		break;
	}

	if(DialogBoxParam(hInst,MAKEINTRESOURCE(DLG_SNAPLOCK),hwnd,
		InputMatrixDlg,(int)dd)==IDOK)
	switch(ActiveProject->EditMode) {
		case EM_MESHEDIT:
			((PMODEL)ActiveProject)->MeshSnap=dd[1].data.fdata;
		break;
		case EM_MAPEDIT:
			((PMODEL)ActiveProject)->MapSnap=dd[1].data.fdata;
		break;
	}

	DestroyDialogData(dd);
}

void ToggleAll(void) {
	vert* v;

	STEPVERTS(ActiveProject,v)
	v->Flags^=TF_MARKED;

	InvalidateView();
}

void LockMarked(void) {
	vert* v;

	STEPVERTS(ActiveProject,v)
	if(v->Flags&TF_MARKED) {
		v->Flags^=VF_LOCKED;
//		CreateVertDelta(v,DT_CHG);
	}


//	AddAtomDelta(ActiveProject);
	InvalidateView();
}

void DetachMarked(void) {
	vert* v;

	STEPVERTS(ActiveProject,v)
	if(v->Flags&TF_MARKED) {
		v->Owner=NULL;
		v->Flags&=~VF_WITHB;
//		CreateVertDelta(v,DT_CHG);
	}
	ClearAll();
//	AddAtomDelta(ActiveProject);
}

void ExtrudeMarked(void) {
	vert *v,*nv;
	edge* e;
	face *f,*nf;

	if(ActiveProject==NULL) return;

	STEPVERTS(ActiveProject,v) v->Copy=NULL;

	STEPEDGES(ActiveProject,e)
	if(e->VertS->Flags & e->VertE->Flags & TF_MARKED)
	if(e->FaceL==NULL || e->FaceR==NULL) {
		f=e->FaceL;
		if(f==NULL) f=e->FaceR;

		v=e->VertS;
		if(v->Copy==NULL) {
			v->Copy=nv=CreateVertex(NULL,0,0,FALSE);
			nv->Spot[X]=v->Spot[X];
			nv->Spot[Y]=v->Spot[Y];
			nv->Spot[Z]=v->Spot[Z];
//			CreateVertDelta(nv,DT_ADD);
		}
		v=e->VertE;
		if(v->Copy==NULL) {
			v->Copy=nv=CreateVertex(NULL,0,0,FALSE);
			nv->Spot[X]=v->Spot[X];
			nv->Spot[Y]=v->Spot[Y];
			nv->Spot[Z]=v->Spot[Z];
//			CreateVertDelta(nv,DT_ADD);
		}
		nf=CreateFace(NULL,e->VertS,e->VertE,e->VertE->Copy,e->VertS->Copy,FALSE);
		if(f) {
			nf->Texture=f->Texture;
			if(f->MapA)
			AddMapping((PMODEL)ActiveProject,nf,TRUE);
//			CreateFaceDelta(nf,DT_ADD);
		}
	}

	STEPVERTS(ActiveProject,v) v->Flags&=~TF_MARKED;

	STEPVERTS(ActiveProject,v)
	if(v->Copy)
	v->Copy->Flags|=TF_MARKED;

//	AddAtomDelta(ActiveProject);
	InvalidateView();
}

void RotateMarked(HWND hwnd,short action) {
	vert* v;
	PDLGX dd;
	float sinx,siny,sinz;
	float cosx,cosy,cosz;
	matrix qm;

	if(ActiveProject==NULL) return;

	dd=(PDLGX)malloc(3*sizeof(DLGX));

	dd[0].item=2;
	dd[1].type=DXT_TTL;
	dd[1].data.tdata="Rotate by";
	dd[2].type=DXT_VEC;
	dd[2].item=EB_VECX;
	dd[2].data.vdata=ActiveProject->UserRotate;

	if(DialogBoxParam(hInst,MAKEINTRESOURCE(DLG_VECTOR),hwnd,
		InputMatrixDlg,(int)dd)==IDOK) {
		switch(action) {
			case EA_EXTRUDE: ExtrudeMarked(); break;
			case EA_COPY: CopyMarked(); break;
		}
		sinx=ActiveProject->UserRotate[X]*DEG2RAD;
		siny=ActiveProject->UserRotate[Y]*DEG2RAD;
		sinz=ActiveProject->UserRotate[Z]*DEG2RAD;
		cosx=(float)cos((double)sinx);
		sinx=(float)sin((double)sinx);
		cosy=(float)cos((double)siny);
		siny=(float)sin((double)siny);
		cosz=(float)cos((double)sinz);
		sinz=(float)sin((double)sinz);

		qm[XX]=cosy*cosz;
		qm[XY]=cosy*sinz;
		qm[XZ]=-siny;

		qm[YX]=sinx*siny*cosz-cosx*sinz;
		qm[YY]=sinx*siny*sinz+cosx*cosz;
		qm[YZ]=sinx*cosy;

		qm[ZX]=cosx*siny*cosz+sinx*sinz;
		qm[ZY]=cosx*siny*sinz-sinx*cosz;
  		qm[ZZ]=cosx*cosy;

		qm[LX]=
		qm[LY]=
		qm[LZ]=0.0;

		STEPVERTS(ActiveProject,v)
		if(v->Flags&TF_MARKED) {
			TransformVector(qm,v->Spot,v->Save);
			v->Spot[X]=v->Save[X];
			v->Spot[Y]=v->Save[Y];
			v->Spot[Z]=v->Save[Z];
//			CreateVertDelta(v,DT_CHG);
		}
		InvalidateView();
//		AddAtomDelta(ActiveProject);
	}

	DestroyDialogData(dd);
}

void ScaleMarked(HWND hwnd,short action) {
	vert *v;
	PDLGX dd;

	if(ActiveProject==NULL) return;

	dd=(PDLGX)malloc(3*sizeof(DLGX));

	dd[0].item=2;
	dd[1].type=DXT_TTL;
	dd[1].data.tdata="Scale by";
	dd[2].type=DXT_VEC;
	dd[2].item=EB_VECX;
	dd[2].data.vdata=ActiveProject->UserScale;

	if(DialogBoxParam(hInst,MAKEINTRESOURCE(DLG_VECTOR),hwnd,
		InputMatrixDlg,(long int)dd)==IDOK) {
		switch(action) {
			case EA_EXTRUDE: ExtrudeMarked(); break;
			case EA_COPY: CopyMarked(); break;
		}
		STEPVERTS(ActiveProject,v)
		if(v->Flags&TF_MARKED) {
			v->Spot[X]*=ActiveProject->UserScale[X];
			v->Spot[Y]*=ActiveProject->UserScale[Y];
			v->Spot[Z]*=ActiveProject->UserScale[Z];
//			CreateVertDelta(v,DT_CHG);
		}
//		AddAtomDelta(ActiveProject);
		InvalidateView();
	}

	DestroyDialogData(dd);
}

void TranslateMarked(HWND hwnd,short action) {
	vert* v;
	PDLGX dd;
//	MATRIX qm;

	if(ActiveProject==NULL) return;

	dd=(PDLGX)malloc(3*sizeof(DLGX));

	dd[0].item=2;
	dd[1].type=DXT_TTL;
	dd[1].data.tdata="Translate by";
	dd[2].type=DXT_VEC;
	dd[2].item=EB_VECX;
	dd[2].data.vdata=ActiveProject->UserTranslate;

	if(DialogBoxParam(hInst,MAKEINTRESOURCE(DLG_VECTOR),hwnd,
		InputMatrixDlg,(int)dd)==IDOK) {
		switch(action) {
			case EA_EXTRUDE: ExtrudeMarked(); break;
			case EA_COPY: CopyMarked(); break;
		}
		STEPVERTS(ActiveProject,v)
		if(v->Flags&TF_MARKED) {
			v->Spot[X]+=ActiveProject->UserTranslate[X];
			v->Spot[Y]+=ActiveProject->UserTranslate[Y];
			v->Spot[Z]+=ActiveProject->UserTranslate[Z];
//			CreateVertDelta(v,DT_CHG);
		}
//		AddAtomDelta(ActiveProject);
		InvalidateView();
	}

	DestroyDialogData(dd);
}

void MatrixMarked(HWND hwnd,short action) {
	short i;
	vert* v;
	PDLGX dd;

	if(ActiveProject==NULL) return;
	dd=(PDLGX)malloc(14*sizeof(DLGX));

	dd[ 0].item=13;
	dd[ 1].type=DXT_TTL;
	dd[ 1].data.tdata="Matrix transform";

	for(i=0;i<12;i++) {
		dd[i+2].type=DXT_NUM;
		dd[i+2].item=EB_VECXX+i;
		dd[i+2].data.fdata=ActiveProject->UserMatrix[i];
	}
	if(DialogBoxParam(hInst,MAKEINTRESOURCE(DLG_MATRIX),hwnd,
		InputMatrixDlg,(int)dd)==IDOK) {
		switch(action) {
			case EA_EXTRUDE: ExtrudeMarked(); break;
			case EA_COPY: CopyMarked(); break;
		}
		for(i=0;i<12;i++) ActiveProject->UserMatrix[i]=dd[i+2].data.fdata;

		STEPVERTS(ActiveProject,v)
		if(v->Flags&TF_MARKED) {
			TransformVector(ActiveProject->UserMatrix,v->Spot,v->Save);
			v->Spot[X]=v->Save[X];
			v->Spot[Y]=v->Save[Y];
			v->Spot[Z]=v->Save[Z];
//			CreateVertDelta(v,DT_CHG);
		}
//		AddAtomDelta(ActiveProject);
		InvalidateView();
	}

	DestroyDialogData(dd);
}

/*
// function to merge marked vertices (I seem to miss when trying to hit them during certain
// operations)
void MergeMarked(void) {
	int vCount;
	VECTOR Sum;
	vert* v,uv;
	vert* va,vb,vc,vd;
	edge* e,ne;
	face* f,nf;

	Sum[X]=0;Sum[Y]=0;Sum[Z]=0;
	vCount=0;

	STEPVERTS(ActiveProject,v)
	if(v->Flags&TF_MARKED) {
		uv=v; // easiest way to get one of them
		vCount++;
		Sum[X]+=v->Spot[X];
		Sum[Y]+=v->Spot[Y];
		Sum[Z]+=v->Spot[Z];
	}
	if(vCount<2) return; // kinda pointless to merge none or only one

	uv->Flags&=~TF_MARKED;
	uv->Spot[X]=Sum[X]/(float)vCount;
	uv->Spot[Y]=Sum[Y]/(float)vCount;
	uv->Spot[Z]=Sum[Z]/(float)vCount;

	STEPEDGES(ActiveProject,e) {
		if(e->VertS->Flags&TF_MARKED) e->VertS=uv;
		if(e->VertE->Flags&TF_MARKED) e->VertE=uv;
	}

	STEPFACES(ActiveProject,f) {
		if(f->VertA->Flags&TF_MARKED) f->VertA=uv;
		if(f->VertB->Flags&TF_MARKED) f->VertB=uv;
		if(f->VertC->Flags&TF_MARKED) f->VertC=uv;
		if(f->VertD!=NULL)
		if(f->VertD->Flags&TF_MARKED) f->VertD=uv;
	}

// if I have other structures that use the hull vertices, have those references replaced as well.
	DeleteVerts(FALSE);	// removes the marked vertices only

// clear face-edge data for all edges that start or end at merge vertex
	STEPEDGES(ActiveProject,e)
	if(e->VertS==e->VertE)
	e->FaceL=e->FaceR=NULL;

	STEPEDGES(ActiveProject,e)
	if(e->VertS==uv || e->VertE==uv)
	if(e->VertS!=e->VertE)
	while(FindDuplicate(NULL,e)) {
		ne=FindDuplicate(NULL,e);
		e->Flags|=ne->Flags;
		ne->VertS=ne->VertE=uv;
	}

	e=FIRSTEDGE(ActiveProject);
	while((edge*)NEXT(e)) {
		ne=(edge*)NEXT(e);
		if(e->VertS==e->VertE)
		DestroyEdge(e,FALSE);
		e=ne;
	}

	f=FIRSTFACE(ActiveProject);
	while((nf=(face*)NEXT(f))==NULL) {
		if(f->VertA==uv || f->VertB==uv || f->VertC==uv || f->VertD==uv) {
			va=f->VertA;
			vb=f->VertB;
			vc=f->VertC;
			vd=f->VertD;
			DestroyFace(f,FALSE);
			CreateFace(NULL,va,vb,vc,vd);
		}
		f=nf;
	}

}
*/

void SmoothEdges(void) {
	edge* e;

	STEPEDGES(ActiveProject,e)
	if(e->VertS->Flags&e->VertE->Flags&TF_MARKED) {
		e->Flags&=~TF_SHARP;
//		CreateEdgeDelta(e,DT_CHG);
	}
	InvalidateView();
//	AddAtomDelta(ActiveProject);
}

void SharpenEdges(void) {
	edge* e;

	STEPEDGES(ActiveProject,e)
	if(e->VertS->Flags&e->VertE->Flags&TF_MARKED) {
		e->Flags|=TF_SHARP;
//		CreateEdgeDelta(e,DT_CHG);
	}
	InvalidateView();
//	AddAtomDelta(ActiveProject);
}

/*
void SmoothVertices(void) {
	vert* v;

	STEPVERTS(ActiveProject,v)
	if(v->Flags&TF_MARKED) {
		v->Flags&=~TF_SHARP;
		RefreshVertex(v);
	}
}

void SharpenVertices(void) {
	vert* v;

	STEPVERTS(ActiveProject,v)
	if(v->Flags&TF_MARKED) {
		v->Flags|=TF_SHARP;
		RefreshVertex(v);
	}
}
*/

void BlankSelection(void) {
	vert* v;

	STEPVERTS(ActiveProject,v)
	v->Select=0;
}

void IncSelection(void) {
	vert* v;

	STEPVERTS(ActiveProject,v)
	if(v->Select)
	v->Select++;
}

void DecSelection(void) {
	vert* v;

	STEPVERTS(ActiveProject,v)
	if(v->Select)
	v->Select--;
}

void LockHighlight(void) {
	switch(HighType) {
	    case HT_VERT:
		InvalidateView();
		HighItem.v->Flags^=VF_LOCKED;
//		CreateVertDelta(HighItem.v,DT_CHG);
//		AddAtomDelta(ActiveProject);
		InvalidateView();
		break;
	}
}

void ToggleHighlight(void) {
	switch(HighType) {
		case HT_EDGE:
			HighItem.e->Flags^=TF_SHARP;
//			CreateEdgeDelta(HighItem.e,DT_CHG);
//			AddAtomDelta((PPROJECT)FindModel((link*)HighItem.e));
			InvalidateView();
		break;
	    case HT_VERT:
		InvalidateView();
		HighItem.v->Flags^=TF_MARKED;
		InvalidateView();
		break;
	}
}

/*
void CopyMarked(PMODEL m) {
	MYPTR mp,op;

	if(m==NULL) m=(PMODEL)ActiveProject;
	if(m==NULL) return;

	STEPVERTS(m,mp.v) mp.v->Copy=NULL;

	STEPVERTS(m,mp.v)
	if(mp.v->Flags&TF_MARKED) {
		op.v=CreateVertex(m,mp.v->pX,mp.v->pY);
		op.v->Spot[X]=mp.v->Spot[X];
		op.v->Spot[Y]=mp.v->Spot[Y];
		op.v->Spot[Z]=mp.v->Spot[Z];
		mp.v->Copy=op.v;
	}

	STEPEDGES(m,mp.e)
	if(mp.e->VertS->Flags & mp.e->VertE->Flags & TF_MARKED) {
		op.e=CreateEdge(m,mp.e->VertS,mp.e->VertE);
		op.e->Flags=mp.e->Flags;
	}

	STEPFACES(m,mp.f) {
		op.f=NULL;
		if(mp.f->VertA->Flags & mp.f->VertB->Flags & mp.f->VertC->Flags & TF_MARKED)
		if(mp.f->VertD) {
			if(mp.f->VertD->Flags & TF_MARKED)
			op.f=CreateFace(m,mp.f->VertA->Copy,mp.f->VertB->Copy,mp.f->VertC->Copy,mp.f->VertD->Copy);
		} else {
			op.f=CreateFace(m,mp.f->VertA->Copy,mp.f->VertB->Copy,mp.f->VertC->Copy,NULL);
		}

		if(op.f) {
			op.f->Flags=mp.f->Flags;
			op.f->Texture=mp.f->Texture;
			if(mp.f->Mapping) {
				op.f->Mapping=CopySurface(mp.f);
			} else {
				op.f->Mapping=NULL;
			}
		}
	}

	STEPVERTS(m,mp.v) mp.v->Flags&=~TF_MARKED;

	STEPVERTS(m,mp.v)
	if(mp.v->Copy) mp.v->Copy->Flags|=TF_MARKED;

	InvalidateView();
}
*/

void AlignMarked(HWND hwnd,int coord) {
	vert *v,*sv;
	PDLGX dd;

	if(ActiveProject==NULL) return;

	if(HighType==HT_VERT) {
		sv=HighItem.v;
	} else {
		sv=(vert*)0;
		STEPVERTS(ActiveProject,v)
		if(v->Flags&TF_MARKED)
		sv=v;
	}

	if(sv==(vert*)0) return;

	dd=(PDLGX)malloc(3*sizeof(DLGX));
	dd[0].item=2;
	dd[1].type=DXT_NUM;
	dd[1].item=EB_VECX;
	dd[1].data.fdata=sv->Spot[coord];
	dd[2].type=DXT_TTL;
	switch(coord) {
		case X:
			dd[2].data.tdata="X coordinate";
		break;
		case Y:
			dd[2].data.tdata="Y coordinate";
		break;
		case Z:
			dd[2].data.tdata="Z coordinate";
		break;
	}
	if(DialogBoxParam(hInst,MAKEINTRESOURCE(DLG_SNAPLOCK),hwnd,
		InputMatrixDlg,(int)dd)==IDOK)
	STEPVERTS(ActiveProject,v)
	if(v->Flags&TF_MARKED) {
		v->Spot[coord]=dd[1].data.fdata;
//		CreateVertDelta(v,DT_CHG);
	}
//	AddAtomDelta(ActiveProject);
	InvalidateView();

	DestroyDialogData(dd);
}

void CopyMarked(void) {
	vert *v,*nv;
	edge *e,*ne;
	face *f,*nf;

	if(ActiveProject==NULL) return;

// copy vertices
	STEPVERTS(ActiveProject,v)
 	if(v->Flags&TF_MARKED) {
		v->Copy=nv=CreateVertex(NULL,0,0,FALSE);
		nv->Spot[X]=v->Spot[X];
		nv->Spot[Y]=v->Spot[Y];
		nv->Spot[Z]=v->Spot[Z];
		nv->Flags=TF_NEW;
//		CreateVertDelta(nv,DT_ADD);
	} else {
		v->Copy=NULL;
		v->Flags&=~TF_NEW;
	}

// copy marked edges
	STEPEDGES(ActiveProject,e)
	if(e->VertS->Flags & e->VertE->Flags & TF_MARKED) {
		ne=CreateEdge(NULL,e->VertS->Copy,e->VertE->Copy,FALSE);
		ne->Flags=e->Flags&TF_SHARP;
//		CreateEdgeDelta(ne,DT_ADD);
	}

// copy marked faces
	STEPFACES(ActiveProject,f) {
		nf=NULL;
		if(f->VertA->Flags & f->VertB->Flags & f->VertC->Flags &TF_MARKED)
		if(f->VertD==NULL) {
			nf=CreateFace(NULL,f->VertA->Copy,f->VertB->Copy,f->VertC->Copy,NULL,
				FALSE);
		} else {
			if(f->VertD->Flags&TF_MARKED)
			nf=CreateFace(NULL,f->VertA->Copy,f->VertB->Copy,f->VertC->Copy,
				f->VertD->Copy,FALSE);
		}
		if(nf) {
			nf->Texture=f->Texture;
			nf->MapA=CopyUVPt(f->MapA);
			nf->MapB=CopyUVPt(f->MapB);
			nf->MapC=CopyUVPt(f->MapC);
			nf->MapD=CopyUVPt(f->MapD);
//			CreateFaceDelta(nf,DT_ADD);
		}
	}

	STEPVERTS(ActiveProject,v) v->Flags&=~TF_MARKED;

	STEPVERTS(ActiveProject,v)
	if(v->Copy)
	v->Copy->Flags|=TF_MARKED;

//	AddAtomDelta(ActiveProject);
	InvalidateView();
}
