#include "data.h"

extern PPROJECT ActiveProject;
extern MYPTR	HighItem;
extern short	HighType;

face* CreateFace(PMODEL m,vert* a, vert* b,vert* c,vert* d,BOOL make_delta) {
	face* f;
	edge *ee,*ef,*eg,*eh;

	if(m==NULL) m=(PMODEL)ActiveProject;

	if(d==NULL) {
		if(a==b || a==c || b==c) return NULL;
	} else {
		if(b==d || a==c) return NULL;
		if(a==b) { b=c; c=d; d=NULL; }
		if(b==c) { c=d; d=NULL; }
		if(c==d) d=NULL;
	}

	ee=FindEdge(m,a,b);
	if(ee!=NULL)
	if(ee->FaceL!=NULL && ee->FaceR!=NULL) return NULL;

	ef=FindEdge(m,b,c);
	if(ef!=NULL)
	if(ef->FaceL!=NULL && ef->FaceR!=NULL) return NULL;

	if(d==NULL) {
		eg=FindEdge(m,c,a);
		if(eg!=NULL)
		if(eg->FaceL!=NULL && eg->FaceR!=NULL) return NULL;
		eh=NULL;
	} else {
		eg=FindEdge(m,c,d);
		if(eg!=NULL)
		if(eg->FaceL!=NULL && eg->FaceR!=NULL) return NULL;

		eh=FindEdge(m,d,a);
		if(eh!=NULL)
		if(eh->FaceL!=NULL && eh->FaceR!=NULL) return NULL;
	}

	f=(face*)malloc(sizeof(face));
	if(f!=NULL) {
		memset(f,0,sizeof(face));
		addTail(&m->Faces,(link*)f);
		((PPROJECT)m)->Flags|=MF_MESHCHANGED;
		f->VertA=a;
		f->VertB=b;
		f->VertC=c;
		f->VertD=d;
//		f->Flags=0;
		f->Index=m->NextFaceIndex++;
		f->Texture=m->defTint;

		if(ee==NULL) ee=CreateEdge(m,a,b,make_delta);
		f->EdgeE=ee;
		if(ee->FaceL==NULL) {
			ee->FaceL=f;
		} else {
			ee->FaceR=f;
		}

		if(ef==NULL) ef=CreateEdge(m,b,c,make_delta);
		f->EdgeF=ef;
		if(ef->FaceL==NULL) {
			ef->FaceL=f;
		} else {
			ef->FaceR=f;
		}

		if(d==NULL) {
			if(eg==NULL) eg=CreateEdge(m,c,a,make_delta);
			f->EdgeG=eg;
			if(eg->FaceL==NULL) {
				eg->FaceL=f;
			} else {
				eg->FaceR=f;
			}
			f->EdgeH=NULL;
		} else {
			if(eg==NULL) eg=CreateEdge(m,c,d,make_delta);
			f->EdgeG=eg;
			if(eg->FaceL==NULL) {
				eg->FaceL=f;
			} else {
				eg->FaceR=f;
			}

			if(eh==NULL) eh=CreateEdge(m,d,a,make_delta);
			f->EdgeH=eh;
			if(eh->FaceL==NULL) {
				eh->FaceL=f;
			} else {
				eh->FaceR=f;
			}
		}
		if(make_delta) CreateFaceDelta(f,DT_ADD);
	}
	return f;
}

PDELT CreateFaceDelta(face* f,TYPE type) {
	PDELT d;
	PMODEL m;

	LSMWORD extra[5];

	m=FindModel((link*)f);

	d=CreateDelta((PPROJECT)m,type,FACEDELT,0);
	if(d==NULL) {
		ReportSystemError("Creation of Delta object failed.");
		return d;
	}
	d->Index=f->Index;

	extra[0].l=f->VertA->Index;
	extra[1].l=f->VertB->Index;
	extra[2].l=f->VertC->Index;
	if(f->VertD) {
		d->ObjectType=QUADDELT;
		extra[3].l=f->VertD->Index;
	} else {
		d->ObjectType=FACEDELT;
	}

	if(type==DT_DEL) return d;

	ExtendDelta(d,(f->VertD==NULL)?3:4,extra);

// extensions for tint
	extra[0].l=TINTATTR;
	extra[1].l=f->Texture->Index;
	ExtendDelta(d,2,extra);

// extensions for uv-mapping
	if(f->MapA) {
		extra[0].l=UVMAPDATA;
		extra[1].l=f->MapA->Index;
		extra[2].l=f->MapB->Index;
		extra[3].l=f->MapC->Index;
		if(f->VertD) {
			extra[4].l=f->MapD->Index;
			ExtendDelta(d,5,extra);
		} else {
			ExtendDelta(d,4,extra);
		}
	}
	return d;
}

void DestroyFace(face* f,BOOL make_delta) {
	edge* e;
	PMODEL m;

	if(f==NULL) return;

	if(HighType==HT_FACE && HighItem.f==f) {
		HighItem.f=NULL;
		HighType=HT_NOTHING;
		InvalidateView();
	}

	if(make_delta) CreateFaceDelta(f,DT_DEL);

	m=FindModel((link*)f);

	STEPEDGES(ActiveProject,e) {
		if(e->FaceL==f) e->FaceL=NULL;
		if(e->FaceR==f) e->FaceR=NULL;
	}
	((PPROJECT)m)->Flags|=MF_MESHCHANGED;

	RemMapping(f,make_delta);

	remLink((link*)f);
	free(f);

	return;
}

void FinishFaceByDelta(PMODEL m,face* f,PDELT d) {
	UINT i;

	if(f==NULL) {
		ReportSystemError("Restored face==NULL");
		return;
	}

	f->Index=d->Index;

	f->VertA=IndexVert(m,d->Data[0].l);
	f->VertB=IndexVert(m,d->Data[1].l);
	f->VertC=IndexVert(m,d->Data[2].l);
	if(d->ObjectType==QUADDELT)
	f->VertD=IndexVert(m,d->Data[3].l);

	i=(f->VertD==NULL)?3:4;
	while(i<d->Len)
	switch(d->Data[i].l) {
		case TINTATTR:
			f->Texture=IndexTint(m,d->Data[i+1].l);
			if(f->Texture==NULL) f->Texture=m->defTint;
			i+=2;
		break;
		case UVMAPDATA:
			f->MapA=IndexUVPt(m,d->Data[i+1].l);
			f->MapB=IndexUVPt(m,d->Data[i+2].l);
			f->MapC=IndexUVPt(m,d->Data[i+3].l);
			i+=4;
			if(f->VertD!=NULL) {
				f->MapD=IndexUVPt(m,d->Data[i].l);
				i++;
			}
		break;
		default: i++;
	}
}

tint *GetFaceTint(face* f) {
	if(f==NULL) return NULL;

	return f->Texture;
}

face* OtherSide(edge* e,face* f) {
	if(e==NULL) return NULL;

	if(f==e->FaceL) {
		return e->FaceR;
	} else {
		return e->FaceL;
	}
}

