#include "data.h"
#include <stdio.h>

extern HWND hwndMain;
extern BOOL EndOfFile;
extern char InputBuffer[64];
extern char FullPathBuffer[256];

extern OPENFILENAME FileData;

void WriteLSMFile(HANDLE fhand,PMODEL m) {
	MYPTR mp;

	WriteLong(fhand,LSM3BLOCK);
	STEPKNOBS(m,mp.k) WriteKnobBlock(fhand,mp.k);
	STEPPARTS(m,mp.p) WritePartBlock(fhand,mp.p);
	STEPVERTS(m,mp.v) WriteVertBlock(fhand,mp.v);
	STEPEDGES(m,mp.e) WriteEdgeBlock(fhand,mp.e);
	STEPTINTS(m,mp.t) WriteTintBlock(fhand,mp.t);
	STEPUVPTS(m,mp.u) WriteUVPtBlock(fhand,mp.u);
	STEPFACES(m,mp.f) WriteFaceBlock(fhand,mp.f);
}

PMODEL ReadLSMFile(HANDLE fhand) {
	long type;
	PMODEL m;

	m=CreateModel();
	if(m!=NULL) {
		DestroyDelta(((PPROJECT)m)->CurrentDelta);
		((PPROJECT)m)->CurrentDelta=NULL;
		DestroyString(m->Project.SaveTitle);
		DestroyString(m->Project.SavePath);
		m->Project.SaveTitle=CloneString(FileData.lpstrFileTitle);
		m->Project.SavePath=CloneString(FileData.lpstrFile);

		type=ReadLong(fhand);

		switch(type) {
			case LSM0BLOCK: ReadLSM0File(fhand,m); break;
			case LSM1BLOCK:
			case LSM2BLOCK:
			case LSM3BLOCK:	ReadLSM1File(fhand,m); break;
			default:
				ReportSystemError("This is not a recongnized file type.");
				return (PMODEL)0;
		}
		BaseParts(m);
		RebuildAllMatrices(m);
	}
	return m;
}

void ReadLSM0File(HANDLE fhand,PMODEL m) {
	long type;

	type=0;
	do {
		if(type==0) type=ReadLong(fhand);
		switch(type) {
			case OLDKNOBBLOCK:	type=ReadKnobBlock(fhand,m);	break;
			case OLDPARTBLOCK:	type=ReadPartBlock(fhand,m);	break;
			case OLDVERTBLOCK:	type=ReadVertBlock(fhand,m);	break;
			case OLDEDGEBLOCK:	type=ReadEdgeBlock(fhand,m);	break;
			case OLDFACEBLOCK:	type=ReadFaceBlock(fhand,m);	break;
			case OLDQUADBLOCK:	type=ReadQuadBlock(fhand,m);	break;
			case OLDTINTBLOCK:	type=ReadTintBlock(fhand,m);	break;
			case OLDUVPTBLOCK:	type=ReadUVPtBlock(fhand,m);	break;
// only listed block types are legal at this level
			default:	type=0;	break;
		}
	} while(EndOfFile==FALSE);
	CalcModelAngles(m);
	AddAtomDelta((PPROJECT)m);
}

void ReadLSM1File(HANDLE fhand,PMODEL m) {
	DWORD count;
	TYPE	type;
	short	len;
	PDELT d;

	while(TRUE) {
		type=ReadShort(fhand);
		if(EndOfFile) {
			CalcModelAngles(m);
			AddAtomDelta((PPROJECT)m);
			return;
		}
		len=ReadShort(fhand);
//		if(len<0) ReportSystemError("Negative block length found in file");
		d=CreateDelta((PPROJECT)m,DT_ADD,type,(int)len);
		switch(type) {
			case EDGEDELT:	d->Index=m->NextEdgeIndex; break;
			case FACEDELT:
			case QUADDELT:	d->Index=m->NextFaceIndex; break;
			case VERTDELT:	d->Index=m->NextVertIndex; break;
			case UVPTDELT:	d->Index=m->NextUVPtIndex; break;
			case TINTDELT:	d->Index=m->NextTintIndex; break;
			case PARTDELT:	d->Index=m->NextPartIndex; break;
			case KNOBDELT:	d->Index=m->NextKnobIndex; break;
		}

		ReadFile(fhand,d->Data,len*sizeof(LSMWORD),&count,NULL);
		RebuildObjectByDelta((PPROJECT)m,d);
  	}
}

void WriteKnobBlock(HANDLE fhand,knob *k) {
	PDELT d;
	d=CreateKnobDelta(k,DT_ADD);
	WriteDelta(fhand,d);
}

long ReadKnobBlock(HANDLE fhand,PMODEL m) {
	knob *k;

	k=CreateKnob(m,0,0,TRUE);
	k->Spot[X]=ReadFloat(fhand);
	k->Spot[Y]=ReadFloat(fhand);
	k->Spot[Z]=ReadFloat(fhand);

	return 0;
}

void WritePartBlock(HANDLE fhand,part* p) {
	PDELT d;

	d=CreatePartDelta(p,DT_ADD);
	WriteDelta(fhand,d);
}

long ReadPartBlock(HANDLE fhand,PMODEL m) {
	knob *kb,*ke,*kr;
	part* p;
	long type;
	float discard;

	kb=IndexKnob(m,ReadLong(fhand));
	ke=IndexKnob(m,ReadLong(fhand));
	kr=IndexKnob(m,ReadLong(fhand));
	p=CreatePart(m,kb,ke,kr,TRUE);
	ke->Owner=p;

	type=0;
	do {
		if(type==0) type=ReadLong(fhand);
		switch(type) {
		case LIMXBLOCK:	// was in former formats, no longer used
			discard=ReadFloat(fhand);
			discard=ReadFloat(fhand);
			type=0;
		break;
		case LIMYBLOCK:	// was in former formats, no longer used
			discard=ReadFloat(fhand);
			discard=ReadFloat(fhand);
			type=0;
		break;
		case LIMZBLOCK:	// was in former formats, no longer used
			discard=ReadFloat(fhand);
			discard=ReadFloat(fhand);
			type=0;
		break;
		case LIMLBLOCK:	// was in former formats, no longer used
			discard=ReadFloat(fhand);
			discard=ReadFloat(fhand);
			type=0;
		break;
		case LOCKXBLOCK:
			p->Flags|=PF_LOCKX;
			type=0;
		break;
		case LOCKYBLOCK:
			p->Flags|=PF_LOCKY;
			type=0;
		break;
		case LOCKZBLOCK:
			p->Flags|=PF_LOCKZ;
			type=0;
		break;
		case LOCKLBLOCK:
			p->Flags|=PF_LOCKL;
			type=0;
		break;
		case NAMEBLOCK:
			ReadPaddedString(fhand,InputBuffer,(int)60);
			InputBuffer[60]=0;
			if(p->Name!=NULL) DestroyString(p->Name);
			p->Name=CloneString(InputBuffer);
			type=0;
		break;
		default:	return type;
		}
	} while(EndOfFile==FALSE);
	return 0;
}

void WriteUVPtBlock(HANDLE fhand,uvpt* u) {
	PDELT d;

	d=CreateUVPtDelta(u,DT_ADD);
	WriteDelta(fhand,d);
}

long ReadUVPtBlock(HANDLE fhand,PMODEL m) {
	uvpt* u;

	u=CreateUVPt(m,TRUE);
	u->Spot[U]=ReadFloat(fhand);
	u->Spot[V]=ReadFloat(fhand);

	return 0;
}

void WriteVertBlock(HANDLE fhand,vert *v) {
	PDELT d;

	d=CreateVertDelta(v,DT_ADD);
	WriteDelta(fhand,d);
}

long ReadVertBlock(HANDLE fhand,PMODEL m) {
	vert *v;
	long type;

	v=CreateVertex(m,0,0,FALSE);
	v->Spot[X]=ReadFloat(fhand);
	v->Spot[Y]=ReadFloat(fhand);
	v->Spot[Z]=ReadFloat(fhand);
	v->Owner=IndexPart(m,ReadLong(fhand));

	type=0;
	do {
		if(type==0) type=ReadLong(fhand);
		switch(type) {
		case OLDFLAG0:
		case ATBASEFLAG:
			v->Flags|=VF_WITHB;
			type=0;
		break;
		case SHARPFLAG:
			v->Flags|=TF_SHARP;
			type=0;
		break;
		default:
			CreateVertDelta(v,DT_ADD);
	  	return type;
		}
	} while(EndOfFile==FALSE);
	CreateVertDelta(v,DT_ADD);
	return 0;
}

void WriteEdgeBlock(HANDLE fhand,edge* e) {
	PDELT d;

	d=CreateEdgeDelta(e,DT_ADD);

	if(e->Flags&TF_SHARP)
	WriteDelta(fhand,d);
}

long ReadEdgeBlock(HANDLE fhand,PMODEL m) {
	edge* e,*te;
	vert *vs,*ve;
	long type;

	vs=IndexVert(m,ReadLong(fhand));
	ve=IndexVert(m,ReadLong(fhand));
	te=FindEdge(m,vs,ve);
	e=GetEdge(m,vs,ve,TRUE);

	type=0;
	do {
		if(type==0) type=ReadLong(fhand);
		switch(type) {
			case OLDFLAG1:
			case SHARPFLAG:
				e->Flags|=TF_SHARP;
				type=0;
			break;
			default:
				if(te==NULL) {
    				CreateEdgeDelta(e,DT_ADD);
   				} else {
    				CreateEdgeDelta(e,DT_CHG);
				}
			return type;
		}
	} while(EndOfFile==FALSE);
	if(te==NULL) {
		CreateEdgeDelta(e,DT_ADD);
	} else {
		CreateEdgeDelta(e,DT_CHG);
	}
	return 0;
}

void WriteFaceBlock(HANDLE fhand,face* f) {
	PDELT d;

	d=CreateFaceDelta(f,DT_ADD);
	WriteDelta(fhand,d);
}

long ReadFaceBlock(HANDLE fhand,PMODEL m) {
	face* f;
	vert *va,*vb,*vc;

	va=IndexVert(m,ReadLong(fhand));
	vb=IndexVert(m,ReadLong(fhand));
	vc=IndexVert(m,ReadLong(fhand));
	if(va==NULL || vb==NULL || vc==NULL)
		ReportSystemError("NULL vert alert!");
	f=CreateFace(m,va,vb,vc,NULL,FALSE);
//	f->Index=m->Faces.Count-1;

	return ReadPaintInfo(fhand,m,f);
}

long ReadQuadBlock(HANDLE fhand,PMODEL m) {
	face* f;
	vert *va,*vb,*vc,*vd;

	va=IndexVert(m,ReadLong(fhand));
	vb=IndexVert(m,ReadLong(fhand));
	vc=IndexVert(m,ReadLong(fhand));
	vd=IndexVert(m,ReadLong(fhand));
	f=CreateFace(m,va,vb,vc,vd,FALSE);
//	f->Index=m->Faces.Count-1;

	return ReadPaintInfo(fhand,m,f);
}

long ReadPaintInfo(HANDLE fhand,PMODEL m,face* f) {
	long type;

	type=0;
	do {
		if(type==0) type=ReadLong(fhand);
		switch(type) {
		case TINTATTR:
			f->Texture=IndexTint(m,ReadLong(fhand));
			type=0;
		break;
		case UVMAPDATA:
//			CreateSurface(m,f,FALSE);
			f->MapA=IndexUVPt(m,ReadLong(fhand));
			f->MapB=IndexUVPt(m,ReadLong(fhand));
			f->MapC=IndexUVPt(m,ReadLong(fhand));
			if(f->VertD)
			f->MapD=IndexUVPt(m,ReadLong(fhand));
			type=0;
		break;
		default: return type;
		}

	} while (EndOfFile==FALSE);
	return type;
}

void WriteTintBlock(HANDLE fhand,tint *t) {
	PDELT d;

	d=CreateTintDelta(t,DT_ADD);
	if(t->Index!=(INDEX)-1)	//	the default texture is never written to file
	WriteDelta(fhand,d);
}

long ReadTintBlock(HANDLE fhand,PMODEL m) {
	tint *t;
	long type;

	t=CreateTint(m,ReadLong(fhand),TRUE);

	type=0;
	do {
		if(type==0) type=ReadLong(fhand);
		switch(type) {
		case NAMEBLOCK:
			ReadPaddedString(fhand,InputBuffer,(int)60);
			DestroyString(t->Name);
			t->Name=CloneString(InputBuffer);
			type=0;
		break;
		case FILEBLOCK:
			ReadPaddedString(fhand,FullPathBuffer,(int)252);
			DestroyString(t->File);
			t->File=CloneString(FullPathBuffer);
			type=0;
		break;
		default: return type;
		}

	} while (EndOfFile==FALSE);
	return type;
}

void WriteDelta(HANDLE fhand,PDELT d) {
	DWORD count;

	WriteShort(fhand,d->ObjectType);
	WriteShort(fhand,d->Len);
	WriteFile(fhand,d->Data,d->Len*sizeof(LSMWORD),&count,NULL);
}

