#include <windows.h>
#include "data.h"
#include "lsmres.h"

extern PPROJECT ActiveProject;

extern HINSTANCE hInst;

extern HWND hwndMain;

extern short HighType;
extern MYPTR HighItem;

tint* CreateTint(PMODEL m,COLORREF c,BOOL make_delta) {
   tint* t;

   if(m==NULL) m=(PMODEL)ActiveProject;

   t=(tint*)malloc(sizeof(tint));

   if(t) {
		memset(t,0,sizeof(tint));
		addTail(&m->Tints,(link*)t);
		t->color=c;
		t->Name=CreateTintName(m);
		t->Index=m->NextTintIndex++;
		if(make_delta) CreateTintDelta(t,DT_ADD);
    }
	return t;
}

PDELT CreateTintDelta(tint* t,TYPE dtype) {
	PDELT d;
	PMODEL m;
	LSMWORD extra[1];

	m=FindModel((link*)t);

	d=CreateDelta((PPROJECT)m,dtype,TINTDELT,0);
	if(d==NULL) {
		ReportSystemError("Creation of Delta object failed.");
		return d;
	}
	d->Index=t->Index;
	if(dtype==DT_DEL) return d;

	extra[0].l=t->color;
	ExtendDelta(d,1,extra);

// extension for name
	if(t->Name) {
		extra[0].l=NAMEBLOCK;
		ExtendDelta(d,1,extra);
		ExtendDelta(d,(strlen(t->Name)+4)>>2,(LSMWORD *)t->Name);
	}
// extension for filename
	if(t->File) {
		extra[0].l=FILEBLOCK;
		ExtendDelta(d,1,extra);
		ExtendDelta(d,(strlen(t->File)+4)>>2,(LSMWORD *)t->File);
	}

	return d;
}

void DestroyTint(tint* t,BOOL make_delta) {
	face* f;
	PMODEL m;

	if(t==NULL) return;

	m=FindModel((link*)t);

	STEPFACES(m,f)
	if(f->Texture==t) {
		f->Texture=((PMODEL)ActiveProject)->defTint;
		if(make_delta) CreateFaceDelta(f,DT_CHG);
	}

	if(t==m->CurTint) m->CurTint=m->defTint;

	if(t->hbm) DeleteObject(t->hbm);

	CreateTintDelta(t,DT_DEL);

	remLink((link*)t);
	DestroyString(t->Name);
	free(t);
}

void FinishTintByDelta(PMODEL m,tint* t,PDELT d) {
	UINT i;

	t->Index=d->Index;
	t->color=d->Data[0].l;

	i=1;
	while(i<d->Len)
	switch(d->Data[i].l) {
		case NAMEBLOCK:
			i++;
			t->Name=CloneString((char*)(d->Data+i));
			i+=((strlen(t->Name)+4)>>2);
		break;
		case FILEBLOCK:
			i++;
			t->File=CloneString((char*)(d->Data+i));
			i+=((strlen(t->File)+4)>>2);
		break;
		default: i++;
	}
	GetTintBitmap(t);
}

void EditFaceTint(void) {
	face *f;

	if(HighType==HT_FACE) {
		f=HighItem.f;
		if(DialogBoxParam(hInst,MAKEINTRESOURCE(DLG_PICKTINT),hwndMain,
			PickTintDlg,(int)f)==IDOK)
		CreateFaceDelta(f,DT_CHG);
	}
}

void BuildTintBitmap(HDC dc,tint* t) {
	BITMAPINFO bmi;
	HBITMAP hbm;
	HBRUSH hbr;
	BITMAP bm;
	HDC pdc,adc;
	RECT dr;
	int nX,nY;
	GLubyte *nbuff;

	if(t==NULL) return;

 	if(t->hbm) DeleteObject(t->hbm);

// create a device context
	pdc=CreateCompatibleDC(dc);
 	memset(&bmi,0,sizeof(BITMAPINFO));
 	bmi.bmiHeader.biSize=sizeof(BITMAPINFOHEADER);
 	bmi.bmiHeader.biPlanes=1;
 	bmi.bmiHeader.biBitCount=24;
	bmi.bmiHeader.biCompression=BI_RGB;

	hbm=(HBITMAP)0;

	if(t->File) { // load the image
		hbm=(HBITMAP)LoadImage(NULL,t->File,IMAGE_BITMAP,0,0,LR_DEFAULTCOLOR
		|LR_CREATEDIBSECTION|LR_LOADFROMFILE);
	}

	if(hbm==(HBITMAP)0) {
// create a DIB section
	 	bmi.bmiHeader.biWidth=128;
	 	bmi.bmiHeader.biHeight=128;

	 	hbm=CreateDIBSection(pdc,&bmi,DIB_RGB_COLORS,(void **)&(t->bmBits),NULL,0);
		SelectObject(pdc,hbm);

		hbr=CreateSolidBrush(t->color);
		SelectObject(pdc,hbr);

		dr.left=dr.top=0;
		dr.right=dr.bottom=128;

		FillRect(pdc,&dr,hbr);
		SetROP2(pdc,R2_NOT);
//  U
		MoveToEx(pdc,56,1,NULL);
		LineTo(pdc,56,33);
		LineTo(pdc,72,33);
		LineTo(pdc,72,0);
//	D
		MoveToEx(pdc,56,126,NULL);
		LineTo(pdc,56,94);
		LineTo(pdc,64,94);
		LineTo(pdc,72,102);
		LineTo(pdc,72,118);
		LineTo(pdc,64,126);
		LineTo(pdc,56,126);
//  L
	    MoveToEx(pdc,1,48,NULL);
	    LineTo(pdc,1,80);
	    LineTo(pdc,18,80);
//  R
		MoveToEx(pdc,110,80,NULL);
		LineTo(pdc,110,48);
		LineTo(pdc,118,48);
		LineTo(pdc,126,56);
		LineTo(pdc,118,64);
		LineTo(pdc,126,72);
		LineTo(pdc,126,81);
		MoveToEx(pdc,111,64,NULL);
		LineTo(pdc,118,64);

		DeleteObject(hbr);
	}
	t->hbm=hbm;

	GetObject(t->hbm,sizeof(BITMAP),&bm);

	t->sX=bm.bmWidth;
	t->sY=bm.bmHeight;

	nX=nearestPowerof2(t->sX);
	nY=nearestPowerof2(t->sY);

	if(nX!=t->sX || nY!=t->sY) {
	 	bmi.bmiHeader.biWidth=nX;
	 	bmi.bmiHeader.biHeight=nY;

	 	hbm=CreateDIBSection(dc,&bmi,DIB_RGB_COLORS,(void **)&nbuff,NULL,0);

		adc=CreateCompatibleDC(dc);
		if(hbm==(HBITMAP)0) ReportSystemError("Bitmap creation failed");
		if(nbuff==0) ReportSystemError("No bits reserved for bitmap");

		SelectObject(pdc,t->hbm);
		SelectObject(adc,hbm);

		SetStretchBltMode(adc,COLORONCOLOR);
		StretchBlt(adc,0,0,nX,nY,pdc,0,0,t->sX,t->sY,SRCCOPY);

		t->bmBits=nbuff;
		DeleteObject(t->hbm);
		t->hbm=hbm;
		DeleteDC(adc);
		t->sX=nX;
		t->sY=nY;
	}
	DeleteDC(pdc);
	DeriveSolid(t);
}

void TintWindow(HWND hwnd,tint* t) {
	RECT trect;
	HDC tdc,sdc;

	if(t==NULL) t=((PMODEL)ActiveProject)->defTint;

	GetClientRect(hwnd,&trect);
	tdc=GetDC(hwnd);
	sdc=CreateCompatibleDC(tdc);

	if(t->hbm==NULL) BuildTintBitmap(tdc,t);

	SelectObject(sdc,t->hbm);

	SetStretchBltMode(tdc,COLORONCOLOR);
	StretchBlt(tdc,0,0,128,128,sdc,0,0,t->sX,t->sY,SRCCOPY);

	trect.left-=2;
	trect.top-=2;
	trect.right=130;
	trect.bottom=130;

	DrawEdge(tdc,&trect,EDGE_SUNKEN,BF_RECT);

	DeleteDC(sdc);
	ReleaseDC(hwnd,tdc);
}

void FillWindow(HWND hwnd,COLORREF cr) {
	RECT trect;
	HBRUSH hbr;
	HDC tdc;

	GetClientRect(hwnd,&trect);
	tdc=GetDC(hwnd);
	hbr=CreateSolidBrush(cr);

	FillRect(tdc,&trect,hbr);

	DrawEdge(tdc,&trect,EDGE_SUNKEN,BF_RECT);

	ReleaseDC(hwnd,tdc);
	DeleteObject(hbr);
}

void CurTintHighFace(void) {
	if(ActiveProject==NULL) return;

	if(HighType!=HT_FACE) return;

	ApplyFaceTint(HighItem.f,((PMODEL)ActiveProject)->CurTint);
}

void ApplyFaceTint(face* f,tint* t) {
	if(f==NULL) return;

	f->Texture=t;
	CreateFaceDelta(f,DT_CHG);
}

void GetTintBitmap(tint* t) {
	BITMAP bm;
	HDC q_dc;

	q_dc=GetDC(NULL);

	if(t->hbm==NULL) {
		q_dc=GetDC(NULL);
		BuildTintBitmap(q_dc,t);
		ReleaseDC(NULL,q_dc);
	}

	GetObject(t->hbm,sizeof(BITMAP),&bm);

	t->bmBits=(GLubyte *)bm.bmBits;
	t->sX=bm.bmWidth;
	t->sY=bm.bmHeight;
}

void DeriveSolid(tint* t) {
	int i,j;
	float r,g,b;
	GLubyte *p;

	if(t->File==NULL) return;

	if(t->bmBits==(GLubyte*)0) GetTintBitmap(t);

	p=t->bmBits;

	r=g=b=0.0;

	for (i=0;i<t->sX;i++)
	for (j=0;j<t->sY;j++) {
		b+=*(p++);
		g+=*(p++);
		r+=*(p++);
	}

	t->color=RGB(r/(float)(t->sX*t->sY),g/(float)(t->sX*t->sY),b/(float)(t->sX*t->sY));
}

